@@ -28,7 +28,6 @@ from tempfile import TemporaryDirectory
 import typing as T
 import uuid
 import textwrap
-import copy
 
 from mesonbuild import mlog
 
@@ -36,8 +35,7 @@ if T.TYPE_CHECKING:
     from .._typing import ImmutableListProtocol
     from ..build import ConfigurationData
     from ..coredata import KeyedOptionDictType, UserOption
-    from ..compilers.compilers import CompilerType
-    from ..interpreterbase import ObjectHolder
+    from ..compilers.compilers import Compiler
 
 FileOrString = T.Union['File', str]
 
@@ -46,7 +44,6 @@ _U = T.TypeVar('_U')
 
 __all__ = [
     'GIT',
-    'an_unpicklable_object',
     'python_command',
     'project_meson_versions',
     'HoldableObject',
@@ -64,6 +61,7 @@ __all__ = [
     'OptionKey',
     'dump_conf_header',
     'OptionOverrideProxy',
+    'OptionProxy',
     'OptionType',
     'OrderedSet',
     'PerMachine',
@@ -266,12 +264,6 @@ def check_direntry_issues(direntry_array: T.Union[T.List[T.Union[str, bytes]], s
                 not pure ASCII. This may cause problems.
                 '''), file=sys.stderr)
 
-
-# Put this in objects that should not get dumped to pickle files
-# by accident.
-import threading
-an_unpicklable_object = threading.Lock()
-
 class HoldableObject(metaclass=abc.ABCMeta):
     ''' Dummy base class for all objects that can be
         held by an interpreter.baseobjects.ObjectHolder '''
@@ -429,6 +421,10 @@ class File(HoldableObject):
             absdir = builddir
         return os.path.join(absdir, self.relative_name())
 
+    @property
+    def suffix(self) -> str:
+        return os.path.splitext(self.fname)[1][1:].lower()
+
     def endswith(self, ending: str) -> bool:
         return self.fname.endswith(ending)
 
@@ -453,7 +449,7 @@ class File(HoldableObject):
         return os.path.join(self.subdir, self.fname)
 
 
-def get_compiler_for_source(compilers: T.Iterable['CompilerType'], src: str) -> 'CompilerType':
+def get_compiler_for_source(compilers: T.Iterable['Compiler'], src: 'FileOrString') -> 'Compiler':
     """Given a set of compilers and a source, find the compiler for that source type."""
     for comp in compilers:
         if comp.can_compile(src):
@@ -461,8 +457,8 @@ def get_compiler_for_source(compilers: T.Iterable['CompilerType'], src: str) ->
     raise MesonException(f'No specified compiler can handle file {src!s}')
 
 
-def classify_unity_sources(compilers: T.Iterable['CompilerType'], sources: T.Iterable[str]) -> T.Dict['CompilerType', T.List[str]]:
-    compsrclist = {}  # type: T.Dict[CompilerType, T.List[str]]
+def classify_unity_sources(compilers: T.Iterable['Compiler'], sources: T.Sequence['FileOrString']) -> T.Dict['Compiler', T.List['FileOrString']]:
+    compsrclist: T.Dict['Compiler', T.List['FileOrString']] = {}
     for src in sources:
         comp = get_compiler_for_source(compilers, src)
         if comp not in compsrclist:
@@ -590,7 +586,6 @@ class PerMachineDefaultable(PerMachine[T.Optional[_T]]):
         return m.default_missing()
 
 
-
 class PerThreeMachineDefaultable(PerMachineDefaultable, PerThreeMachine[T.Optional[_T]]):
     """Extends `PerThreeMachine` with the ability to default from `None`s.
     """
@@ -1060,7 +1055,7 @@ if is_windows():
             if c == '\\':
                 num_backslashes += 1
             else:
-                if c == '"' and not (num_backslashes % 2):
+                if c == '"' and not num_backslashes % 2:
                     # unescaped quote, eat it
                     arg += (num_backslashes // 2) * '\\'
                     num_quotes += 1
@@ -1097,7 +1092,7 @@ def join_args(args: T.Iterable[str]) -> str:
 
 
 def do_replacement(regex: T.Pattern[str], line: str, variable_format: str,
-                   confdata: 'ConfigurationData') -> T.Tuple[str, T.Set[str]]:
+                   confdata: T.Union[T.Dict[str, T.Tuple[str, T.Optional[str]]], 'ConfigurationData']) -> T.Tuple[str, T.Set[str]]:
     missing_variables = set()  # type: T.Set[str]
     if variable_format == 'cmake':
         start_tag = '${'
@@ -1120,7 +1115,7 @@ def do_replacement(regex: T.Pattern[str], line: str, variable_format: str,
             varname = match.group(1)
             var_str = ''
             if varname in confdata:
-                (var, desc) = confdata.get(varname)
+                var, _ = confdata.get(varname)
                 if isinstance(var, str):
                     var_str = var
                 elif isinstance(var, int):
@@ -1261,6 +1256,8 @@ def dump_conf_header(ofilename: str, cdata: 'ConfigurationData', output_format:
     elif output_format == 'nasm':
         prelude = CONF_NASM_PRELUDE
         prefix = '%'
+    else:
+        raise MesonBugException(f'Undefined output_format: "{output_format}"')
 
     ofilename_tmp = ofilename + '~'
     with open(ofilename_tmp, 'w', encoding='utf-8') as ofile:
@@ -1301,7 +1298,6 @@ def replace_if_different(dst: str, dst_tmp: str) -> None:
         os.unlink(dst_tmp)
 
 
-
 def listify(item: T.Any, flatten: bool = True) -> T.List[T.Any]:
     '''
     Returns a list with all args embedded in a list if they are not a list.
@@ -1384,7 +1380,7 @@ def partition(pred: T.Callable[[_T], object], iterable: T.Iterable[_T]) -> T.Tup
 def Popen_safe(args: T.List[str], write: T.Optional[str] = None,
                stdout: T.Union[T.TextIO, T.BinaryIO, int] = subprocess.PIPE,
                stderr: T.Union[T.TextIO, T.BinaryIO, int] = subprocess.PIPE,
-               **kwargs: T.Any) -> T.Tuple[subprocess.Popen, str, str]:
+               **kwargs: T.Any) -> T.Tuple['subprocess.Popen[str]', str, str]:
     import locale
     encoding = locale.getpreferredencoding()
     # Redirect stdin to DEVNULL otherwise the command run by us here might mess
@@ -1407,7 +1403,7 @@ def Popen_safe(args: T.List[str], write: T.Optional[str] = None,
 def Popen_safe_legacy(args: T.List[str], write: T.Optional[str] = None,
                       stdout: T.Union[T.TextIO, T.BinaryIO, int] = subprocess.PIPE,
                       stderr: T.Union[T.TextIO, T.BinaryIO, int] = subprocess.PIPE,
-                      **kwargs: T.Any) -> T.Tuple[subprocess.Popen, str, str]:
+                      **kwargs: T.Any) -> T.Tuple['subprocess.Popen[str]', str, str]:
     p = subprocess.Popen(args, universal_newlines=False, close_fds=False,
                          stdout=stdout, stderr=stderr, **kwargs)
     input_ = None  # type: T.Optional[bytes]
@@ -1443,7 +1439,7 @@ def iter_regexin_iter(regexiter: T.Iterable[str], initer: T.Iterable[str]) -> T.
     return None
 
 
-def _substitute_values_check_errors(command: T.List[str], values: T.Dict[str, str]) -> None:
+def _substitute_values_check_errors(command: T.List[str], values: T.Dict[str, T.Union[str, T.List[str]]]) -> None:
     # Error checking
     inregex = ['@INPUT([0-9]+)?@', '@PLAINNAME@', '@BASENAME@']  # type: T.List[str]
     outregex = ['@OUTPUT([0-9]+)?@', '@OUTDIR@']                 # type: T.List[str]
@@ -1471,8 +1467,7 @@ def _substitute_values_check_errors(command: T.List[str], values: T.Dict[str, st
         # Error out if any output-derived templates are present in the command
         match = iter_regexin_iter(outregex, command)
         if match:
-            m = 'Command cannot have {!r} since there are no outputs'
-            raise MesonException(m.format(match))
+            raise MesonException(f'Command cannot have {match!r} since there are no outputs')
     else:
         # Error out if an invalid @OUTPUTnn@ template was specified
         for each in command:
@@ -1484,7 +1479,7 @@ def _substitute_values_check_errors(command: T.List[str], values: T.Dict[str, st
                 raise MesonException(m.format(match2.group(), len(values['@OUTPUT@'])))
 
 
-def substitute_values(command: T.List[str], values: T.Dict[str, str]) -> T.List[str]:
+def substitute_values(command: T.List[str], values: T.Dict[str, T.Union[str, T.List[str]]]) -> T.List[str]:
     '''
     Substitute the template strings in the @values dict into the list of
     strings @command and return a new list. For a full list of the templates,
@@ -1493,14 +1488,29 @@ def substitute_values(command: T.List[str], values: T.Dict[str, str]) -> T.List[
     If multiple inputs/outputs are given in the @values dictionary, we
     substitute @INPUT@ and @OUTPUT@ only if they are the entire string, not
     just a part of it, and in that case we substitute *all* of them.
+
+    The typing of this function is difficult, as only @OUTPUT@ and @INPUT@ can
+    be lists, everything else is a string. However, TypeDict cannot represent
+    this, as you can have optional keys, but not extra keys. We end up just
+    having to us asserts to convince type checkers that this is okay.
+
+    https://github.com/python/mypy/issues/4617
     '''
+
+    def replace(m: T.Match[str]) -> str:
+        v = values[m.group(0)]
+        assert isinstance(v, str), 'for mypy'
+        return v
+
     # Error checking
     _substitute_values_check_errors(command, values)
+
     # Substitution
     outcmd = []  # type: T.List[str]
     rx_keys = [re.escape(key) for key in values if key not in ('@INPUT@', '@OUTPUT@')]
     value_rx = re.compile('|'.join(rx_keys)) if rx_keys else None
     for vv in command:
+        more: T.Optional[str] = None
         if not isinstance(vv, str):
             outcmd.append(vv)
         elif '@INPUT@' in vv:
@@ -1521,15 +1531,22 @@ def substitute_values(command: T.List[str], values: T.Dict[str, str]) -> T.List[
             else:
                 raise MesonException("Command has '@OUTPUT@' as part of a "
                                      "string and more than one output file")
+
         # Append values that are exactly a template string.
         # This is faster than a string replace.
         elif vv in values:
-            outcmd.append(values[vv])
+            o = values[vv]
+            assert isinstance(o, str), 'for mypy'
+            more = o
         # Substitute everything else with replacement
         elif value_rx:
-            outcmd.append(value_rx.sub(lambda m: values[m.group(0)], vv))
+            more = value_rx.sub(replace, vv)
         else:
-            outcmd.append(vv)
+            more = vv
+
+        if more is not None:
+            outcmd.append(more)
+
     return outcmd
 
 
@@ -1714,7 +1731,7 @@ class OrderedSet(T.MutableSet[_T]):
         return 'OrderedSet()'
 
     def __reversed__(self) -> T.Iterator[_T]:
-        # Mypy is complaining that sets cant be reversed, which is true for
+        # Mypy is complaining that sets can't be reversed, which is true for
         # unordered sets, but this is an ordered, set so reverse() makes sense.
         return reversed(self.__container.keys())  # type: ignore
 
@@ -1902,6 +1919,16 @@ def run_once(func: T.Callable[..., _T]) -> T.Callable[..., _T]:
     return wrapper
 
 
+class OptionProxy(T.Generic[_T]):
+    def __init__(self, value: _T, choices: T.Optional[T.List[str]] = None):
+        self.value = value
+        self.choices = choices
+
+    def set_value(self, v: _T) -> None:
+        # XXX: should this be an error
+        self.value = v
+
+
 class OptionOverrideProxy(collections.abc.MutableMapping):
 
     '''Mimic an option list but transparently override selected option
@@ -1917,16 +1944,15 @@ class OptionOverrideProxy(collections.abc.MutableMapping):
         for o in options:
             self.options.update(o)
 
-    def __getitem__(self, key: 'OptionKey') -> T.Union['UserOption']:
+    def __getitem__(self, key: 'OptionKey') -> T.Union['UserOption', OptionProxy]:
         if key in self.options:
             opt = self.options[key]
             if key in self.overrides:
-                opt = copy.copy(opt)
-                opt.set_value(self.overrides[key])
+                return OptionProxy(opt.validate_value(self.overrides[key]), getattr(opt, 'choices', None))
             return opt
         raise KeyError('Option not found', key)
 
-    def __setitem__(self, key: 'OptionKey', value: T.Union['UserOption']) -> None:
+    def __setitem__(self, key: 'OptionKey', value: T.Union['UserOption', OptionProxy]) -> None:
         self.overrides[key] = value.value
 
     def __delitem__(self, key: 'OptionKey') -> None:
@@ -1942,15 +1968,15 @@ class OptionOverrideProxy(collections.abc.MutableMapping):
         return OptionOverrideProxy(self.overrides.copy(), self.options.copy())
 
 
-class OptionType(enum.Enum):
+class OptionType(enum.IntEnum):
 
     """Enum used to specify what kind of argument a thing is."""
 
     BUILTIN = 0
-    BASE = 1
-    COMPILER = 2
-    PROJECT = 3
-    BACKEND = 4
+    BACKEND = 1
+    BASE = 2
+    COMPILER = 3
+    PROJECT = 4
 
 # This is copied from coredata. There is no way to share this, because this
 # is used in the OptionKey constructor, and the coredata lists are
@@ -1998,7 +2024,7 @@ def _classify_argument(key: 'OptionKey') -> OptionType:
         return OptionType.BASE
     elif key.lang is not None:
         return OptionType.COMPILER
-    elif key.name in _BUILTIN_NAMES:
+    elif key.name in _BUILTIN_NAMES or key.module:
         return OptionType.BUILTIN
     elif key.name.startswith('backend_'):
         assert key.machine is MachineChoice.HOST, str(key)
@@ -2018,7 +2044,7 @@ class OptionKey:
     internally easier to reason about and produce.
     """
 
-    __slots__ = ['name', 'subproject', 'machine', 'lang', '_hash', 'type']
+    __slots__ = ['name', 'subproject', 'machine', 'lang', '_hash', 'type', 'module']
 
     name: str
     subproject: str
@@ -2026,10 +2052,13 @@ class OptionKey:
     lang: T.Optional[str]
     _hash: int
     type: OptionType
+    module: T.Optional[str]
 
     def __init__(self, name: str, subproject: str = '',
                  machine: MachineChoice = MachineChoice.HOST,
-                 lang: T.Optional[str] = None, _type: T.Optional[OptionType] = None):
+                 lang: T.Optional[str] = None,
+                 module: T.Optional[str] = None,
+                 _type: T.Optional[OptionType] = None):
         # the _type option to the constructor is kinda private. We want to be
         # able tos ave the state and avoid the lookup function when
         # pickling/unpickling, but we need to be able to calculate it when
@@ -2038,7 +2067,8 @@ class OptionKey:
         object.__setattr__(self, 'subproject', subproject)
         object.__setattr__(self, 'machine', machine)
         object.__setattr__(self, 'lang', lang)
-        object.__setattr__(self, '_hash', hash((name, subproject, machine, lang)))
+        object.__setattr__(self, 'module', module)
+        object.__setattr__(self, '_hash', hash((name, subproject, machine, lang, module)))
         if _type is None:
             _type = _classify_argument(self)
         object.__setattr__(self, 'type', _type)
@@ -2053,6 +2083,7 @@ class OptionKey:
             'machine': self.machine,
             'lang': self.lang,
             '_type': self.type,
+            'module': self.module,
         }
 
     def __setstate__(self, state: T.Dict[str, T.Any]) -> None:
@@ -2069,22 +2100,17 @@ class OptionKey:
     def __hash__(self) -> int:
         return self._hash
 
+    def _to_tuple(self) -> T.Tuple[str, OptionType, str, str, MachineChoice, str]:
+        return (self.subproject, self.type, self.lang or '', self.module or '', self.machine, self.name)
+
     def __eq__(self, other: object) -> bool:
         if isinstance(other, OptionKey):
-            return (
-                self.name == other.name and
-                self.subproject == other.subproject and
-                self.machine is other.machine and
-                self.lang == other.lang)
+            return self._to_tuple() == other._to_tuple()
         return NotImplemented
 
     def __lt__(self, other: object) -> bool:
         if isinstance(other, OptionKey):
-            return (
-                self.name < other.name and
-                self.subproject < other.subproject and
-                self.machine < other.machine and
-                self.lang < other.lang)
+            return self._to_tuple() < other._to_tuple()
         return NotImplemented
 
     def __str__(self) -> str:
@@ -2093,6 +2119,8 @@ class OptionKey:
             out = f'{self.lang}_{out}'
         if self.machine is MachineChoice.BUILD:
             out = f'build.{out}'
+        if self.module:
+            out = f'{self.module}.{out}'
         if self.subproject:
             out = f'{self.subproject}:{out}'
         return out
@@ -2112,12 +2140,16 @@ class OptionKey:
         except ValueError:
             subproject, raw2 = '', raw
 
-        if raw2.startswith('build.'):
-            raw3 = raw2.split('.', 1)[1]
-            for_machine = MachineChoice.BUILD
-        else:
+        module = None
+        for_machine = MachineChoice.HOST
+        try:
+            prefix, raw3 = raw2.split('.')
+            if prefix == 'build':
+                for_machine = MachineChoice.BUILD
+            else:
+                module = prefix
+        except ValueError:
             raw3 = raw2
-            for_machine = MachineChoice.HOST
 
         from ..compilers import all_languages
         if any(raw3.startswith(f'{l}_') for l in all_languages):
@@ -2125,12 +2157,13 @@ class OptionKey:
         else:
             lang, opt = None, raw3
         assert ':' not in opt
-        assert 'build.' not in opt
+        assert '.' not in opt
 
-        return cls(opt, subproject, for_machine, lang)
+        return cls(opt, subproject, for_machine, lang, module)
 
     def evolve(self, name: T.Optional[str] = None, subproject: T.Optional[str] = None,
-               machine: T.Optional[MachineChoice] = None, lang: T.Optional[str] = '') -> 'OptionKey':
+               machine: T.Optional[MachineChoice] = None, lang: T.Optional[str] = '',
+               module: T.Optional[str] = '') -> 'OptionKey':
         """Create a new copy of this key, but with alterted members.
 
         For example:
@@ -2146,6 +2179,7 @@ class OptionKey:
             subproject if subproject is not None else self.subproject,
             machine if machine is not None else self.machine,
             lang if lang != '' else self.lang,
+            module if module != '' else self.module
         )
 
     def as_root(self) -> 'OptionKey':
