@@ -39,6 +39,48 @@ def gen_guid():
     '''
     return str(uuid.uuid4()).upper()
 
+def get_all_modules_from_dir(dirname):
+    '''
+    Get all modules required for Meson build MSI package
+    from directories.
+    '''
+    modname = os.path.basename(dirname)
+    modules = [os.path.splitext(os.path.split(x)[1])[0] for x in glob(os.path.join(dirname, '*'))]
+    modules = ['mesonbuild.' + modname + '.' + x for x in modules if not x.startswith('_')]
+    return modules
+
+def get_more_modules():
+    '''
+        Getter for missing Modules.
+        Python packagers want to be minimal and only copy the things
+        that they can see that being used. They are blind to many things.
+    '''
+    return ['distutils.archive_util',
+            'distutils.cmd',
+            'distutils.config',
+            'distutils.core',
+            'distutils.debug',
+            'distutils.dep_util',
+            'distutils.dir_util',
+            'distutils.dist',
+            'distutils.errors',
+            'distutils.extension',
+            'distutils.fancy_getopt',
+            'distutils.file_util',
+            'distutils.spawn',
+            'distutils.util',
+            'distutils.version',
+            'distutils.command.build_ext',
+            'distutils.command.build',
+            'filecmp',
+            ]
+
+def get_modules():
+    modules = get_all_modules_from_dir('mesonbuild/modules')
+    modules += get_all_modules_from_dir('mesonbuild/scripts')
+    modules += get_more_modules()
+    return modules
+
 class Node:
     '''
        Node to hold path and directory values
@@ -67,7 +109,7 @@ class Node:
 
 class PackageGenerator:
     '''
-       Package generator for MSI pacakges
+       Package generator for MSI packages
     '''
 
     def __init__(self):
@@ -107,44 +149,6 @@ class PackageGenerator:
         for s_d in self.staging_dirs:
             self.feature_components[s_d] = []
 
-    @staticmethod
-    def get_all_modules_from_dir(dirname):
-        '''
-           Get all modules required for Meson build MSI package
-           from directories.
-        '''
-        modname = os.path.basename(dirname)
-        modules = [os.path.splitext(os.path.split(x)[1])[0] for x in glob(os.path.join(dirname, '*'))]
-        modules = ['mesonbuild.' + modname + '.' + x for x in modules if not x.startswith('_')]
-        return modules
-
-    @staticmethod
-    def get_more_modules():
-        '''
-           Getter for missing Modules.
-           Python packagers want to be minimal and only copy the things
-           that they can see that being used. They are blind to many things.
-        '''
-        return ['distutils.archive_util',
-                'distutils.cmd',
-                'distutils.config',
-                'distutils.core',
-                'distutils.debug',
-                'distutils.dep_util',
-                'distutils.dir_util',
-                'distutils.dist',
-                'distutils.errors',
-                'distutils.extension',
-                'distutils.fancy_getopt',
-                'distutils.file_util',
-                'distutils.spawn',
-                'distutils.util',
-                'distutils.version',
-                'distutils.command.build_ext',
-                'distutils.command.build',
-                'filecmp',
-                ]
-
     def build_dist(self):
         '''
            Build dist file from PyInstaller info
@@ -153,9 +157,7 @@ class PackageGenerator:
             if os.path.exists(sdir):
                 shutil.rmtree(sdir)
         main_stage, ninja_stage = self.staging_dirs
-        modules = self.get_all_modules_from_dir('mesonbuild/modules')
-        modules += self.get_all_modules_from_dir('mesonbuild/scripts')
-        modules += self.get_more_modules()
+        modules = get_modules()
 
         pyinstaller = shutil.which('pyinstaller')
         if not pyinstaller:
@@ -171,9 +173,12 @@ class PackageGenerator:
                       pyinstaller_tmpdir]
         for m in modules:
             pyinst_cmd += ['--hidden-import', m]
+        # https://github.com/pyinstaller/pyinstaller/issues/5693
+        pyinst_cmd += ['--exclude-module', '_bootlocale']
         pyinst_cmd += ['meson.py']
         subprocess.check_call(pyinst_cmd)
         shutil.move(pyinstaller_tmpdir + '/meson', main_stage)
+        self.del_infodirs(main_stage)
         if not os.path.exists(os.path.join(main_stage, 'meson.exe')):
             sys.exit('Meson exe missing from staging dir.')
         os.mkdir(ninja_stage)
@@ -181,6 +186,14 @@ class PackageGenerator:
         if not os.path.exists(os.path.join(ninja_stage, 'ninja.exe')):
             sys.exit('Ninja exe missing from staging dir.')
 
+    def del_infodirs(self, dirname):
+        # Starting with 3.9.something there are some
+        # extra metadatadirs that have a hyphen in their
+        # file names. This is a forbidden character in WiX
+        # filenames so delete them.
+        for d in glob(os.path.join(dirname, '*-info')):
+            shutil.rmtree(d)
+
     def generate_files(self):
         '''
            Generate package files for MSI installer package
